#include "compgeom.h"

#include <math.h>

#ifndef ABS
#define ABS(x) (((x) < 0)? -(x): (x))
#endif

double CompGeom::ABSTOL = 1e-7;

//-------------------------------------------------------------------------
void 
CompGeom::pickLineCode( double x, double y,
                        double xmin, double xmax, double ymin, double ymax,
                        int cod[4] )
{
 cod[0] = x < xmin;
 cod[1] = x > xmax;
 cod[2] = y < ymin;
 cod[3] = y > ymax;
}
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
bool 
CompGeom::isPointCloseToSegment( Point _p1, Point _p2, Point _p, Point _tol )
{
 double x1 = _p1.getX();
 double y1 = _p1.getY();
 double x2 = _p2.getX();
 double y2 = _p2.getY();

 // Define atraction window.
 double xmin = _p.getX() - _tol.getX();
 double xmax = _p.getX() + _tol.getX();
 double ymin = _p.getY() - _tol.getY();
 double ymax = _p.getY() + _tol.getY();

 int cod1[4], cod2[4];
 int i;

 // Check whether point 'p' touch given line segment 'p1'-'p2'
 // using given tolerance values.
 pickLineCode( x2, y2, xmin, xmax, ymin, ymax, cod2 );
 do
 {
  pickLineCode( x1, y1, xmin, xmax, ymin, ymax, cod1 );
  for( i = 0; i < 4; i++ )
  {
   if( cod1[i] && cod2[i] )     // Check of trivial non-pick
    break;
  }
  if( i != 4 )
   break;

  // move point 1 to the limit of attraction window
  if( cod1[0] )
  {
   y1 += (xmin - x1) * (y2 - y1) / (x2 - x1);
   x1 = xmin;
  }
  else if( cod1[1] )
  {
   y1 += (xmax - x1) * (y2 - y1) / (x2 - x1);
   x1 = xmax;
  }
  else if( cod1[2] )
  {
   x1 += (ymin - y1) * (x2 - x1) / (y2 - y1);
   y1 = ymin;
  }
  else if( cod1[3] )
  {
   x1 += (ymax - y1) * (x2 - x1) / (y2 - y1);
   y1 = ymax;
  }
  else
  {
   return true;
  }
 } while( 1 );

 return false;
}
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
double 
CompGeom::getClosestPointLine( Point _p1, Point _p2, Point _p,
                               Point& _pC, double& _t )
{
 Point v12 = _p2 - _p1;
 Point v1p = _p  - _p1;

 _t = dotprod( v12, v1p ) / sizesquare(v12);
 _pC = _p1 + _t * v12;

 double dist = size(_p-_pC);
 return dist;
}
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
double 
CompGeom::getClosestPointSegment( Point _p1, Point _p2, Point _p,
                                  Point& _pC, double& _t )
{
 Point v12 = _p2 - _p1;
 Point v1p = _p  - _p1;

 _t = dotprod( v12, v1p ) / sizesquare(v12);

 if( ABS(_t) < ABSTOL )
 {
  _pC = _p1;
  _t = 0.0;
 }
 else if( ABS(_t-1.0) < ABSTOL )
 {
  _pC = _p2;
  _t = 1.0;
 }
 else
 {
  _pC = _p1 + _t * v12;
 }

 double dist = size(_p-_pC);
 return dist;
}
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
bool 
CompGeom::checkCollinearSegments( Point _p1, Point _p2,
                                  Point _p3, Point _p4 )
{
// Computed oriented areas for points p1-p2-p3 and for points p1-p2-p4
 double area123 = orient2d(_p1,_p2,_p3);
 double area124 = orient2d(_p1,_p2,_p4);

// Check for collinear segments
 if( (ABS(area123) < ABSTOL) && (ABS(area124) < ABSTOL) )
 {
  return true;
 }
 return false;
}
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
int 
CompGeom::computeSegmentSegmentIntersection( Point _p1, Point _p2,
                                             Point _p3, Point _p4,
                                             Point& _pi,
                                             double& _t12, double& _t34 )
{
 Point A = _p2 - _p1;
 Point B = _p3 - _p4;

 // Discard intesection if second segment is located to the left (_l) or 
 // to the right (_r) of horizontal bounding box limits of first segment.
 double x12_l, x12_r;
 double x34_l, x34_r;
 x12_l = (_p1.getX() < _p2.getX()) ? _p1.getX() : _p2.getX();
 x12_r = (_p1.getX() > _p2.getX()) ? _p1.getX() : _p2.getX();
 x34_l = (_p3.getX() < _p4.getX()) ? _p3.getX() : _p4.getX();
 x34_r = (_p3.getX() > _p4.getX()) ? _p3.getX() : _p4.getX();
 if( (x12_r+ABSTOL) < x34_l || x34_r < (x12_l-ABSTOL) )
 {
  return DO_NOT_INTERSECT;
 }

 // Discard intesection if second segment is located below 
 // bottom (_b) or above top (_t) of bounding box of first segment.
 double y12_b, y12_t;
 double y34_b, y34_t;
 y12_b = (_p1.getY() < _p2.getY()) ? _p1.getY() : _p2.getY();
 y12_t = (_p1.getY() > _p2.getY()) ? _p1.getY() : _p2.getY();
 y34_b = (_p3.getY() < _p4.getY()) ? _p3.getY() : _p4.getY();
 y34_t = (_p3.getY() > _p4.getY()) ? _p3.getY() : _p4.getY();
 if( (y12_t+ABSTOL) < y34_b || y34_t < (y12_b-ABSTOL) )
 {
  return DO_NOT_INTERSECT;
 }

// Computed oriented areas for points p1-p2-p3 and for points p1-p2-p4
 double area123 = orient2d(_p1,_p2,_p3);
 double area124 = orient2d(_p1,_p2,_p4);

// Check for collinear segments
/*** COMPLETE AQUI (COMPGEOM): 01 ***/




/*** COMPLETE AQUI (COMPGEOM): 01 ***/

 // Check for parallel segments (not necessary since check below takes it)
 if( ABS(area123-area124) < ABSTOL )
 {
  return PARALLEL;
 }

 // Check for second segment on the same side of first segment
/*** COMPLETE AQUI (COMPGEOM): 02 ***/





/*** COMPLETE AQUI (COMPGEOM): 02 ***/

 // Check for point p3 touching segment p1-p2
/*** COMPLETE AQUI (COMPGEOM): 03 ***/























/*** COMPLETE AQUI (COMPGEOM): 03 ***/

 // Check for point p4 touching segment p1-p2
/*** COMPLETE AQUI (COMPGEOM): 04 ***/























/*** COMPLETE AQUI (COMPGEOM): 04 ***/

 // Computed oriented areas for points p3-p4-p1 and for points p3-p4-p2
 double area341 = orient2d(_p3,_p4,_p1);
 double area342 = orient2d(_p3,_p4,_p2);

 // Check for first segment on the same side of second segment
 /*** COMPLETE AQUI (COMPGEOM): 05 ***/





 /*** COMPLETE AQUI (COMPGEOM): 05 ***/

 // Check for point p1 touching segment p3-p4
/*** COMPLETE AQUI (COMPGEOM): 06 ***/























/*** COMPLETE AQUI (COMPGEOM): 06 ***/

 // Check for point p2 touching segment p3-p4
/*** COMPLETE AQUI (COMPGEOM): 07 ***/























/*** COMPLETE AQUI (COMPGEOM): 07 ***/

 // When get to this point, there is an intersection point of the
 // two segments. Compute parametric values of intersection point
 // along each segment.
/*** COMPLETE AQUI (COMPGEOM): 08 ***/


/*** COMPLETE AQUI (COMPGEOM): 08 ***/

 // Compute intersection point (there are two equivalent options)
// Point v12 = _p2 - _p1;
// _pi = _p1 + _t12 * v12;
 Point v34 = _p4 - _p3;
 _pi = _p3 + _t34 * v34;
 return DO_INTERSECT;
}
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
bool 
CompGeom::isCounterClockwisePolygon( vector<Point>& _poly )
{
 double area = 0.0; // twice the enclosed polygon area

 // Compute area assuming that polygon is in counter-clockwise order
 for( int i = 2; i < _poly.size(); i++ ) 
 {
  area += orient2d( _poly[0], _poly[i-1], _poly[i] );
 }

 // If area is greater than zero, then polygon is in counter-clockwise 
 // order, otherwise it is in clockwise order.
 if( area > 0.0 )
  return true;
 return false;
}
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
bool 
CompGeom::isPointInPolygon( vector<Point>& _poly, Point _p )
{
 double  x = _p.getX();
 double  y = _p.getY();
 int     n = _poly.size();               // number of polygon points
 int     ni = 0;                         // number of intersections

 for( int i = 0; i < n; i++ )
 {
  Point p1 = _poly[i];                   // first point of current line segment
  Point p2 = _poly[(i+1)%n];             // second point of current line segment
/*** COMPLETE AQUI (COMPGEOM): 09 ***/































































/*** COMPLETE AQUI (COMPGEOM): 09 ***/
 }

 // If number of intersections is odd, point is inside polygon.
 if( (ni % 2) > 0 )
  return true;

 // If number of intersections if even, point is outside polygon.
 return false;
}
//-------------------------------------------------------------------------
